import os
import glob
import subprocess
from filelock import FileLock
from omegaconf import OmegaConf
from pprint import pprint
import shutil

from file_golem._file_io_helper import _does_path_exist, \
    _initialize_load_save_extension_dicts
from file_golem.file_datatypes import FileDatatypes, FilePathEntries, AbstractDatatype,SpecialDataArgs
from file_golem.config_datatype import Config
from file_golem.class_handler import ClassHandler
from datetime import datetime


class FileIO:
    DATA_DIRECTORY_NAME = 'data_directory_name'
    PROJECT_BASE_DIRECTORY = 'base_data_directory'
    FILELOCK_TIMEOUT_KEY = 'filelock_timeout'
    MAX_INDEXED_DIRECTORY_SIZE = 'max_indexed_directory_size'
    SYSTEM_TRANSFERS = 'system_transfers'
    PROJECT_SRC_ROOT = 'project_src_root'
    HAS_DUPLICATE_MODULE_NAMES = 'has_duplicate_module_names'
    SYSTEM_IDENTITY_FILE = 'system_identity_file'
    def __init__(self, system_config = None, system_config_path = None,is_debug=False, system_transfer=None):
        
        self.system_config = system_config
        self.system_config_path = system_config_path
        self.system_transfer = system_transfer
        #self.transfer_path_prefix = []

        self.is_debug = is_debug
        self.lengths_cache = {}
        self.config_cache = {}
        self.load_system_config()

        self.base_data_directory_name = self.system_config.get(self.DATA_DIRECTORY_NAME, 'data')

        self.class_handler = ClassHandler(
            self.system_config.get(self.PROJECT_SRC_ROOT, 'src'),
            self.system_config.get(self.HAS_DUPLICATE_MODULE_NAMES, False),
        )


        
        self.load_function_dict, \
        self.save_function_dict, \
        self.file_extension_dict, = _initialize_load_save_extension_dicts(self.system_config[SpecialDataArgs.SUPPORTED_FILETYPES.value])



    def load_system_config(self):
        if self.system_config_path is None:
            raise Exception('You must pass either a system_config or system_config_path to the FileIO constructor')
        
        self.load_function_dict, \
        self.save_function_dict, \
        self.file_extension_dict = _initialize_load_save_extension_dicts([FileDatatypes.OMEGA_CONF.value])

        #self.system_identity = None
        #self.system_paths_dict = {}
        self.system_config = self.load_config(self.system_config_path)

        ####Extract system identity and paths from the system config
        system_identity_file = self.system_config.get(self.SYSTEM_IDENTITY_FILE, '.system_identity.txt')
        with open(system_identity_file, 'r') as system_identity_file:
            self.system_identity = system_identity_file.read().strip()
        if self.is_debug:
            print(f'System Identity: {self.system_identity}')

        system_paths_dict = self.system_config.get(SpecialDataArgs.SYSTEM_PATHS.value, {})
            
        if self.system_transfer is None:
            self.absolute_path_prefix = system_paths_dict.get(self.system_identity,{}
                ).get(self.PROJECT_BASE_DIRECTORY, None).split(os.sep)
            if self.absolute_path_prefix is None:
                raise Exception(f'Base data directory not found in system config for system identity {self.system_identity}. Please specify it under {SpecialDataArgs.SYSTEM_PATHS.value}:{self.system_identity}:{self.PROJECT_BASE_DIRECTORY}')
            
        else:
            self.absolute_path_prefix = system_paths_dict.get(self.system_identity,{}).get(
                self.SYSTEM_TRANSFERS,{}).get(self.system_transfer,None).split(os.sep)
            if self.absolute_path_prefix is None:
                raise Exception(f'Base data directory not found in system config for transfer {self.system_identity} to {self.system_transfer}. Please specify it under {SpecialDataArgs.SYSTEM_PATHS.value}:{self.system_identity}:{self.SYSTEM_TRANSFERS}:{self.system_transfer}:{self.system_transfer}')

            

    def load_config(self,config_path):
        config = self._load_config_recursive(config_path,[])
        return config
    

    def _load_config_recursive(self,config_path,ancestry):
        if config_path in self.config_cache:
            return self.config_cache[config_path]
        if config_path in ancestry:
            raise Exception(f'Cycle detected in config inheritance: {ancestry}')
        
        config =self.load_data(Config, data_args = {
            Config.CONFIG_NAME: config_path
        })

        if self.is_debug:
            print('Loaded config: ', config_path),
            pprint(OmegaConf.to_yaml(config))

        defaults_list = config.get(SpecialDataArgs.DEFAULTS.value,[])
        for default_config_path in defaults_list:
            default_config = self._load_config_recursive(default_config_path,ancestry+[config_path])
            config = OmegaConf.merge(default_config,config)
        
        self.config_cache[config_path] = config
        return config

    def _config_to_tuple_fn(self,config_path):
        cwd = os.getcwd()
        base_cwd = os.path.basename(cwd)
        path_parts = config_path.split(os.sep)
        indices = [i for i, part in enumerate(path_parts) if part == base_cwd]

        if len(indices) > 1:
            print(f'WARNING: config path: {config_path} has multiple instances of {base_cwd}. This may cause issues with config loading.'
                    'It is strongly recommended to have only one instance of the base directory name in the config path.')
        elif len(indices) == 1:                
            path_parts = path_parts[indices[-1]+1:]
        config_file_extension = f'.{self.file_extension_dict[FileDatatypes.OMEGA_CONF]}'
        if path_parts[-1].endswith(config_file_extension):
            path_parts[-1] = path_parts[-1][:-len(config_file_extension)]
        return tuple(path_parts)
    

    def fetch_subconfig(self, config, subconfig_keys):
        if isinstance(config, str):
            config  = self.load_config(config)

        subconfig = config
        for key in subconfig_keys:
            if key not in subconfig:
                return {}
            subconfig = subconfig[key]
            if isinstance(subconfig, str):
                subconfig = self.load_config(subconfig)

        
        return subconfig
    

    def fetch_config_field(self,config, subconfig_keys = [],is_required=True):
        subconfig = self.fetch_subconfig(config,subconfig_keys[:-1])
        if not subconfig:
            if is_required:
                raise Exception(f'No subconfig found for keys {subconfig_keys} in config {config}')
            else:
                return
            
        field = subconfig.get(subconfig_keys[-1],None)

        if field is None:
            if is_required:
                raise Exception(f'No field found for keys {subconfig_keys} in config {config}')
            else:
                return None
        return field
    

    def fetch_class_from_config(self,config,subconfig_keys = [],is_required=True):
        model_class = self.fetch_config_field(config, subconfig_keys=subconfig_keys,is_required=is_required)
        if model_class is None:
            if is_required:
                raise Exception(f'No class found for keys {subconfig_keys} in config {config}')
            else:
                return None
        return self.fetch_class(model_class)

    def fetch_class(self,model_class):
        return self.class_handler._locate_class(model_class)



    def load_data(self,data_type, data_args={}):
        data_path = self.get_data_path(data_type,data_args)

        return self._load_data_core(data_type, data_path)
    
    def _load_data_core(self,data_type, data_path):
        if self.is_debug:
            print('Loading data from',data_path)
        load_function = self.load_function_dict[data_type.FILE_DATATYPE]
        data = load_function(data_path)
        return data
    
    def save_data(self,data_type, data_args):
        data_path = self.get_data_path(data_type,data_args)
        self.create_directory(data_path)
        save_function = self.save_function_dict[data_type.FILE_DATATYPE]
        data = data_args[AbstractDatatype.DATA]
        if self.is_debug:
            print('Saving data to',data_path)
        save_function(data,data_path)

    def create_directory(self,path):
        dir_path = os.path.dirname(path)
        if len(dir_path) == 0:
            return
        if self.is_debug:
            if not os.path.exists(dir_path):
                print(f'Creating directory: {dir_path}')
        os.makedirs(dir_path,exist_ok=True)


    def transfer_data(self,source_datatype,target_datatype, source_data_args, target_data_args):
        source_path = self.get_data_path(source_datatype, data_args=source_data_args)
        target_path = self.get_data_path(target_datatype, data_args=target_data_args)
        self.create_directory(target_path)
        if self.is_debug:
            print(f'Overwriting {target_path} with data at {source_path}')
        shutil.copy(source_path, target_path)



    def delete_data(self,data_type, data_args):
        for file_path in self.get_file_iterator(data_type, data_args = data_args):
            if os.path.isdir(file_path):
                for root, dirs, files in os.walk(file_path, topdown=False):
                    for name in files:
                        os.remove(os.path.join(root, name))
                    for name in dirs:
                        os.rmdir(os.path.join(root, name))
                os.rmdir(file_path)
            else:
                os.remove(file_path)


    def get_data_path(self,datatype,data_args,is_lock_file=False, is_from=None, is_to=None):

        #Build the relative path
        relative_path_list = self._build_relative_path_list(datatype,data_args)
        if datatype.FILE_DATATYPE is not FileDatatypes.EMPTY:
            if datatype.FILE_DATATYPE not in self.file_extension_dict:
                raise Exception(f'File datatype {datatype.FILE_DATATYPE.value} not found in file extension dict. Please check the supported filetypes in the system config.')
            file_extenion = self.file_extension_dict[datatype.FILE_DATATYPE]
            relative_path_list[-1] += f'.{file_extenion}'
        
        if is_lock_file:
            relative_path_list[-1] += '.lock'

        path_prefix = self._get_path_prefix(datatype, is_from, is_to)


        complete_path =  os.path.join( *path_prefix, *relative_path_list)
        if not complete_path.startswith(os.sep):
            complete_path = os.sep + complete_path

        complete_path = self._resolve_symlinks(complete_path)
        #return complete_path

        normed_path = os.path.normpath(complete_path)
        return normed_path
    

    def _resolve_symlinks(self, path):
        if self.system_transfer is None:
            return path
        if self.is_debug:
            print(f'Due to system transfer, resolving symlinks for path: {path}')
        symlink_set = set()
        path = self._resolve_symlinks_recursive(path,symlink_set)
        return path


    def _resolve_symlinks_recursive(self,path,symlink_set):
        path_components = path.split(os.sep)
        #for i in range(len(path_components)-1):
        for i in range(len(path_components)):
            partial_path = '/'+ os.path.join(*path_components[:i+1])
            if os.path.islink(partial_path):
                if partial_path in symlink_set:
                    raise Exception(f'Cycle detected in symlinks: {partial_path} already visited')
                symlink_set.add(partial_path)


                symlink_target = os.readlink(partial_path)
                if self.is_debug:
                    print(f'Found symlink: {partial_path} to {symlink_target}')
                if os.path.isabs(symlink_target):
                    transfer_system_base_directory = self.system_config.get(
                        SpecialDataArgs.SYSTEM_PATHS.value, {}).get(
                        self.system_transfer, {}).get(
                        self.PROJECT_BASE_DIRECTORY, None)
                    relative_path = os.path.relpath(symlink_target, start=transfer_system_base_directory)

                else:
                    raise Exception(f'Symlink target {symlink_target} is not an absolute path. In this case, behavior is unknown. possibly set relative path to symlink target.')
                    relative_path = symlink_target

                unsymlinked_path = os.path.join(*self.absolute_path_prefix,relative_path,*path_components[i+1:])
                if not unsymlinked_path.startswith(os.sep):
                    unsymlinked_path = os.sep + unsymlinked_path
                return self._resolve_symlinks_recursive(unsymlinked_path, symlink_set)

        return path



    def _get_path_prefix(self,datatype,is_from, is_to):
        if datatype is Config:
            return os.getcwd().split(os.sep) #Config is always assumed to be in the current working directory
        
        #prefix = []
        prefix = self.absolute_path_prefix

        if (is_from is not None) and (is_to is not None):
            raise Exception('cannot specify both is_from and is_to')
        elif (is_from is not None) or (is_to is not None):
            shortest_path = self.get_entry_point_path(is_from,is_to)
            prefix += shortest_path.split(os.sep)

        return prefix

        # raise Exception('fix here')

        # if self.system_transfer is not None:
        #     prefix += self.transfer_path_prefix        
        # return prefix
    
    def atomic_operation(self,datatype,data_args,atomic_function,new_data):
        data_lock_path = self.get_data_path(datatype, data_args=data_args,is_lock_file=True)
        if self.is_debug:
            print(f'Creating filelock at {data_lock_path}')
        #timeout = self.system_config[self.FILELOCK_TIMEOUT_KEY]
        timeout = self.system_config.get(self.FILELOCK_TIMEOUT_KEY, 60)
        filelock = FileLock(data_lock_path, timeout=timeout)
        filelock.acquire()
        if self.is_file_present(datatype, data_args=data_args):
            data = self.load_data(datatype,data_args=data_args)
        else:
            data = None
        modified_data = atomic_function(data,new_data)
        data_args[AbstractDatatype.DATA] = modified_data
        self.save_data(datatype, data_args)

        os.remove(data_lock_path)
        filelock.release()
        if self.is_debug:
            print(f'Released and deleted filelock at {data_lock_path}')
        

    def get_file_iterator(self, data_type, data_args):
        partial_path = self.get_data_path(data_type,data_args)
        for file in glob.glob(partial_path,recursive=True):
            yield file
        
    def get_data_iterator(self, data_type, data_args):
        for file_path in self.get_file_iterator(data_type, data_args):
            data = self._load_data_core(data_type, file_path)
            yield data

    def _build_relative_path_list(self,datatype,data_args):
        relative_path_list = []
        for entry in datatype.RELATIVE_PATH_TUPLE:
            if entry == FilePathEntries.BASE_DATA_DIRECTORY:
                relative_path_list.append(self.base_data_directory_name)
            elif entry == FilePathEntries.IDX_ENTRY:
                idx = data_args[AbstractDatatype.IDX]
                if idx == FilePathEntries.OPEN_ENTRY:
                    relative_path_list.append('**')
                else:
                    relative_path_list.append(str(idx))
            elif entry == FilePathEntries.MODULATED_IDX_ENTRY:
                max_idx = self.system_config.get(self.MAX_INDEXED_DIRECTORY_SIZE, 10000)
                raise Exception('Modulated idx entry is not implemented yet')
            elif entry == FilePathEntries.CONFIG_ENTRY:
                config_name = data_args[AbstractDatatype.CONFIG_NAME]
                if config_name is None:
                    raise Exception(f'Config name not found in data args: {data_args}')
                if config_name == FilePathEntries.OPEN_ENTRY:
                    relative_path_list.append('**')
                else:
                    relative_config_tuple = self._config_to_tuple_fn(config_name)
                    relative_path_list.extend(relative_config_tuple)
            elif entry == FilePathEntries.TIMESTAMP_CONFIG_ENTRY:
                config_name = data_args[AbstractDatatype.CONFIG_NAME]
                if config_name is None:
                    raise Exception(f'Config name not found in data args: {data_args}')
                if config_name == FilePathEntries.OPEN_ENTRY:
                    relative_path_list.append('*')
                else:
                    relative_config_tuple = self._config_to_tuple_fn(config_name)
                    timestamp = datetime.now().strftime('%m-%d_%H:%M:%S')
                    relative_path_list.append(f'{relative_config_tuple[-1]}_{timestamp}')
            elif entry == FilePathEntries.DATA_TYPE_ENTRY:
                data_type_entry = datatype.__class__.__name__
                relative_path_list.append(data_type_entry)
            elif isinstance(entry, dict):
                if len(entry) != 1:
                    raise Exception(f'Dictionary entry in RELATIVE_PATH_TUPLE must have exactly one key-value pair, but got: {entry}')
                key, value = next(iter(entry.items()))
                if key == FilePathEntries.DATA_ARG_ENTRY:
                    field = data_args[value]
                    if field == FilePathEntries.OPEN_ENTRY:
                        relative_path_list.append('*')
                    else:
                        relative_path_list.append(str(field))
                elif key == FilePathEntries.CUSTOM_LOGIC:
                    custom_fn = getattr(datatype,value)
                    data_args[SpecialDataArgs.DATA_IO]=self
                    data_args[SpecialDataArgs.DATA_TYPE]=datatype
                    custom_values = custom_fn(data_args)

                    if custom_values is FilePathEntries.OPEN_ENTRY:
                        custom_values = ['*']
                    if isinstance(custom_values, str):
                        custom_values = [custom_values]
                    relative_path_list.extend(custom_values)
                elif key == FilePathEntries.ATTRIBUTE_ENTRY:
                    attribute = getattr(datatype,value)
                    relative_path_list.append(str(attribute))
                elif key == FilePathEntries.ZERO_PADDED_IDX_ENTRY:
                    idx = data_args[AbstractDatatype.IDX]
                    if idx == FilePathEntries.OPEN_ENTRY:
                        relative_path_list.append('*')
                    else:
                        zero_padded_idx = str(idx).zfill(value)
                        relative_path_list.append(zero_padded_idx)
                else:
                    raise Exception(f'Invalid key in dictionary entry in RELATIVE_PATH_TUPLE: {key}')
            elif isinstance(entry, str):
                relative_path_list.append(entry)
            else:
                raise Exception(f'Invalid entry in RELATIVE_PATH_TUPLE: {entry}')
        return relative_path_list


    def retrieve_data_args(self,data_type,partial_data_args, data_path):
        #TODO: FIX TO BE GENERAL
        relative_path_list = data_path.split(os.sep)

        epoch = relative_path_list[-1].split('.')[0]
        missing_data_args = {
            'epoch': epoch
        }
        return missing_data_args

    def get_datatype_length(self,datatype, data_args={}):
        if datatype in self.lengths_cache:
            length = self.lengths_cache[datatype]
        else:
            if datatype.LENGTH == -1:
                data_args[SpecialDataArgs.DATA_IO]=self
                data_args[SpecialDataArgs.DATA_TYPE]=datatype
                length = datatype._custom_length_logic(data_args)
            else:
                length = datatype.LENGTH
            self.lengths_cache[datatype] = length
        return length    

        
    def _get_config_directory(self):
        config_directory_key = 'config_directory'
        if config_directory_key in self.system_config:
            return self.system_config[config_directory_key]
        else:
            raise Exception(f'{config_directory_key} found in system config ({self.system_config_path})')
    
    def is_file_present(self,data_type,data_args):
        data_path = self.get_data_path(data_type,data_args)
        if self.is_debug:
            print('Checking if file is present at',data_path)
        is_file_present = _does_path_exist(data_path)
        return is_file_present
    

    def run_system_command(self,command):
        if self.is_debug:
            print('Running terminal command:',command)
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, shell=True)
        for line in process.stdout:
            print(line, end='')
        process.wait()

    def execute_python_command(self,entry_point_name, args_dict):
        cd_path = self.get_entry_point_path(is_to=entry_point_name)
        conda_path = self.get_conda_path(entry_point_name)
        python_entry_file = self.system_config.get(
            SpecialDataArgs.SYSTEM_PATHS.value, {}).get(
            self.system_identity, {}).get(
            SpecialDataArgs.EXTERNAL_CALLS.value, {}).get(
            entry_point_name, {}).get(
            SpecialDataArgs.PYTHON_ENTRY_POINT.value, None)
        if python_entry_file is None:
            raise Exception(f'Python entry point for {entry_point_name} not found in system config for system identity {self.system_identity}')
        args_string = ''
        for key, value in args_dict.items():
            if value is None:
                args_string += f' --{key}'
            else:
                args_string += f' --{key} {value}'
        cmmd_string = (
            f'cd {cd_path} && {conda_path} {python_entry_file}.py {args_string}'
        )
        self.run_system_command(cmmd_string)

    def get_conda_path(self,entry_point_name):
        conda_path = self.system_config.get(
            SpecialDataArgs.SYSTEM_PATHS.value, {}).get(
            self.system_identity, {}).get(
            SpecialDataArgs.EXTERNAL_CALLS.value, {}).get(
            entry_point_name, {}).get(
            SpecialDataArgs.CONDA_PATH.value, None)
        
        conda_path = os.path.join(conda_path, 'bin', 'python')
        if conda_path is None:
            raise Exception(f'Conda path {entry_point_name} not found in system config for system identity {self.system_identity}')
        return conda_path
    
    def get_entry_point_path(self,is_from=None, is_to=None):
        #TODO: Should deprecate in favor of using abbsolute paths 
        if (is_from is not None) and (is_to is not None):
            raise Exception('cannot specify both is_from and is_to')
        elif is_from is None and is_to is None:
            raise Exception('must specify either is_from or is_to')
        
        entry_point_name = is_from if is_from is not None else is_to
        entry_point_path = self.system_config.get(
            SpecialDataArgs.SYSTEM_PATHS.value, {}).get(
            self.system_identity, {}).get(
            SpecialDataArgs.EXTERNAL_CALLS.value, {}).get(
            entry_point_name, {}).get(
            SpecialDataArgs.ENTRY_POINT.value, None)
        
        if entry_point_path is None:
            raise Exception(f'Entry point {entry_point_name} not found in system config for system identity {self.system_identity}')
        
        cwd = os.getcwd()
        if is_to is not None:
            relpath = os.path.relpath(entry_point_path, cwd)
        else:
            relpath = os.path.relpath(cwd, entry_point_path)
        

        return relpath


from enum import Enum

class FileDatatypes(Enum):
    MATPLOTLIB = 'matplotlib'
    PANDAS = 'pandas'
    JSON = 'json'
    TEXT = 'text'
    TORCH = 'torch'
    PICKLE = 'pickle'
    PNG = 'png'
    PDF = 'pdf'
    NUMPY = 'numpy'
    OMEGA_CONF = 'omega_conf'
    EMPTY = 'empty'
    SHELL = 'shell'
    SLURM_SCRIPT = 'slurm_script'
    SLURM_OUTPUT_STD = 'slurm_output_std'
    SLURM_OUTPUT_ERR = 'slurm_output_err'
    JPEG = 'jpeg'
    JPEG_BOLD = 'jpeg_bold'



class SpecialDataArgs(Enum):
    DATA_IO = 'data_io'
    DATA_TYPE = 'data_type'
    SUPPORTED_FILETYPES = 'supported_filetypes'
    DEFAULTS = 'defaults'
    SYSTEM_PATHS = 'system_paths'
    IGNORE_PRIMARY_DATA_PREFIX_LIST = 'ignore_primary_data_prefix'
    EXTERNAL_CALLS = 'external_calls'
    ENTRY_POINT = 'entry_point'
    CONDA_PATH = 'conda_path'
    PYTHON_ENTRY_POINT = 'python_entry_point'
    OVERLEAF_SUBDIR = 'overleaf_subdir'
    OVERLEAF_MAPPINGS = 'overleaf_mappings'




class FilePathEntries(Enum):
    BASE_DATA_DIRECTORY = 'base_data_directory'
    IDX_ENTRY = 'idx_entry'
    MODULATED_IDX_ENTRY = 'modulated_idx_entry'
    ZERO_PADDED_IDX_ENTRY = 'zero_padded_idx_entry' #Use for zero padded idx entries
    ATTRIBUTE_ENTRY = 'attribute_entry'
    DATA_TYPE_ENTRY = 'data_type_entry'
    OPEN_ENTRY = 'open_entry' #Use as a place holder when iterating over files
    CONFIG_ENTRY = 'config_entry' #Use for parsing config title. This ensures that the seperators are handled correctly
    TIMESTAMP_CONFIG_ENTRY = 'timestamp_config_entry' #Use primarily for slurm outputs
    DATA_ARG_ENTRY = 'data_arg_entry' #Use as a dict key for processing the tuple
    CUSTOM_LOGIC = 'custom_logic' #call into a function to get the entry value


class AbstractDatatype:
    DATA = 'data'
    IDX = 'idx'
    FILE_DATATYPE = None #Assign a value from the FileDatatypes enum here
    CONFIG_NAME = 'config_name'
    LENGTH = -1
    RELATIVE_PATH_TUPLE = (
        FilePathEntries.BASE_DATA_DIRECTORY,
    )

    @staticmethod
    def _custom_length_logic(data_args):
        datatype = data_args[SpecialDataArgs.DATA_TYPE]
        raise Exception(f'Custom length logic (the static method _custom_length_logic(data_args)) not implemented for datatype {datatype.__name__}.')


import os
import ast
import pydoc


class ClassHandler:
    def __init__(self,project_src_root,has_duplicate_module_names):
        self.class_module_dict = {}
        self.has_duplicate_module_names = has_duplicate_module_names
        if self.has_duplicate_module_names:
            return
        current_working_directory = os.getcwd()
        project_path = os.path.join(current_working_directory, project_src_root)

        """Find all classes in a Python project."""
        class_file_paths_dict = {}
        for root, _, files in os.walk(project_path):
            for file in files:
                if file.endswith(".py"):  # Only process Python files
                    file_path = os.path.join(root, file)
                    relative_path = os.path.relpath(file_path, current_working_directory)
                    try:
                        """Find all class names in a Python file."""
                        with open(file_path, "r", encoding="utf-8") as file:
                            tree = ast.parse(file.read(), filename=file_path)

                            for node in ast.walk(tree):
                                if isinstance(node, ast.ClassDef):
                                    class_name = node.name
                                    if class_name in class_file_paths_dict.keys():

                                        print(f"WARNING: Duplicate class name '{class_name}' found in file {relative_path} and {class_file_paths_dict[class_name]}")
                                        print("Duplicate class names can cause issues with module resolution.")
                                        print("Please ensure unique names across files to enable module short names.")
                                        print(f"Or, set has_duplicate_module_names to True in the system config")
                                    #print(f"Found class {node.name} in file {file_path}")
                                    else:
                                        class_file_paths_dict[class_name] = relative_path
                                        module_name = os.path.splitext(relative_path)[0].replace(os.sep, '.')+'.'+class_name
                                        self.class_module_dict[class_name] = module_name
                    except Exception as e:
                        print(f"Error parsing {file_path}: {e}")


    def _locate_class(self,object_class):
        if object_class in self.class_module_dict:
            object_class = self.class_module_dict[object_class]
        model_type = pydoc.locate(object_class)
        if model_type is None:
            raise Exception(f'Could not find model class {object_class}')
        return model_type





from file_golem.file_datatypes import FileDatatypes, FilePathEntries, AbstractDatatype

class Config(AbstractDatatype):
    FILE_DATATYPE = FileDatatypes.OMEGA_CONF
    GRID_IDX = 'grid_idx'

    RELATIVE_PATH_TUPLE = (
        FilePathEntries.CONFIG_ENTRY,
        {FilePathEntries.CUSTOM_LOGIC: '_grid_search_suffix'}
    )
    
    @staticmethod
    def _grid_search_suffix(data_args):
        if not (Config.GRID_IDX in data_args):
            return ()
        
        idx = '_'.join(map(str, data_args[Config.GRID_IDX]))
        return tuple([idx])



import os
from file_golem.file_datatypes import FileDatatypes

############### FILE IO ####################
def _does_path_exist(path):
    return os.path.exists(path)


def _initialize_load_save_extension_dicts(supported_datatypes):
    load_function_dict = {}
    save_function_dict = {}
    file_extension_dict = {}

    can_load_all = (len(supported_datatypes) == 0)

    if FileDatatypes.JPEG.value in supported_datatypes or can_load_all:
        from PIL import Image
        def _load_jpeg(path):
            with Image.open(path) as img:
                return img.convert('RGB')
            
        def _save_jpeg(data,path):
            data.save(path, format='JPEG')


        load_function_dict[FileDatatypes.JPEG] = _load_jpeg
        save_function_dict[FileDatatypes.JPEG] = _save_jpeg
        file_extension_dict[FileDatatypes.JPEG] = 'jpg'

    if FileDatatypes.JPEG_BOLD.value in supported_datatypes or can_load_all:
        from PIL import Image, ImageDraw, ImageFont
        def _load_jpeg_bold(path):
            with Image.open(path) as img:
                return img.convert('RGB')
        
        def _save_jpeg_bold(data,path):
            data.save(path, format='JPEG')

        load_function_dict[FileDatatypes.JPEG_BOLD] = _load_jpeg_bold
        save_function_dict[FileDatatypes.JPEG_BOLD] = _save_jpeg_bold
        file_extension_dict[FileDatatypes.JPEG_BOLD] = 'JPEG'




    if (FileDatatypes.OMEGA_CONF.value in supported_datatypes) or can_load_all:

        from omegaconf import OmegaConf
        def _load_omega_conf(path):
            config = OmegaConf.load(path)
            return config

        def _save_omega_conf(data,path):
            OmegaConf.save(data,path)


        load_function_dict[FileDatatypes.OMEGA_CONF] = _load_omega_conf
        save_function_dict[FileDatatypes.OMEGA_CONF] = _save_omega_conf
        file_extension_dict[FileDatatypes.OMEGA_CONF] = 'yaml'

    if (FileDatatypes.MATPLOTLIB.value) in supported_datatypes or can_load_all:
        import matplotlib.pyplot as plt

        def _save_matplotlib(data,path):
            plt.savefig(path,bbox_inches='tight', pad_inches =0.0,format='pdf')

        save_function_dict[FileDatatypes.MATPLOTLIB] = _save_matplotlib
        file_extension_dict[FileDatatypes.MATPLOTLIB] = 'pdf'


    if (FileDatatypes.NUMPY.value in supported_datatypes) or can_load_all:
        import numpy as np
        def _load_np(path):
            return np.load(path)
        
        def _save_np(data,path):
            np.save(path,data)

        load_function_dict[FileDatatypes.NUMPY] = _load_np
        save_function_dict[FileDatatypes.NUMPY] = _save_np
        file_extension_dict[FileDatatypes.NUMPY] = 'npy'

    if (FileDatatypes.PANDAS.value in supported_datatypes) or can_load_all:
        import pandas as pd
        def _save_pd(data,path):
            data.to_csv(path)
        
        def _load_pd(path):
            return pd.read_csv(path)
        
        load_function_dict[FileDatatypes.PANDAS] = _load_pd
        save_function_dict[FileDatatypes.PANDAS] = _save_pd
        file_extension_dict[FileDatatypes.PANDAS] = 'csv'
    
    if (FileDatatypes.JSON.value in supported_datatypes) or can_load_all:
        import json
        def _load_json(path):
            with open(path, 'r') as f:
                return json.load(f)
            
        def _save_json(json_data,path):
            with open(path, 'w') as f:
                json.dump(json_data, f, indent=4)
        
        load_function_dict[FileDatatypes.JSON] = _load_json
        save_function_dict[FileDatatypes.JSON] = _save_json
        file_extension_dict[FileDatatypes.JSON] = 'json'

    if (FileDatatypes.TEXT.value in supported_datatypes) or can_load_all:
        def _load_txt(file_path):
            with open(file_path, 'r') as file:
                content = file.read()
            print("CAM WUZ HERE: ", content)
            return content

        def _save_txt(data, file_path):
            with open(file_path, 'w') as file:
                file.write(data)

        load_function_dict[FileDatatypes.TEXT] = _load_txt
        save_function_dict[FileDatatypes.TEXT] = _save_txt
        file_extension_dict[FileDatatypes.TEXT] = 'txt'

    if (FileDatatypes.TORCH.value in supported_datatypes) or can_load_all:
        import torch
        def _save_torch(data,path):
            torch.save(data,path)
        def _load_torch(path,weights_only=False):
            if torch.cuda.is_available():
                return torch.load(path,weights_only=weights_only)
            else:
                return torch.load(path,weights_only=weights_only,map_location=torch.device('cpu'))
        
        load_function_dict[FileDatatypes.TORCH] = _load_torch
        save_function_dict[FileDatatypes.TORCH] = _save_torch
        file_extension_dict[FileDatatypes.TORCH] = 'pt'

    if (FileDatatypes.PICKLE.value in supported_datatypes) or can_load_all:
        import pickle
        def _save_pickle(data,path):
            with open(path, 'wb') as f:
                pickle.dump(data, f)

        def _load_pickle(path):
            with open(path, 'rb') as f:
                return pickle.load(f)
            
        load_function_dict[FileDatatypes.PICKLE] = _load_pickle
        save_function_dict[FileDatatypes.PICKLE] = _save_pickle
        file_extension_dict[FileDatatypes.PICKLE] = 'pkl'

    if (FileDatatypes.PNG.value in supported_datatypes) or can_load_all:
        import imageio
        def _load_png(path):
            return imageio.imread(path)
        def _save_png(data,path):
            imageio.imwrite(path, data)
            # with open(path, 'wb') as img_file:
            #     img_file.write(data)
        load_function_dict[FileDatatypes.PNG] = _load_png
        save_function_dict[FileDatatypes.PNG] = _save_png
        file_extension_dict[FileDatatypes.PNG] = 'png'

    if (FileDatatypes.PDF.value in supported_datatypes) or can_load_all:
        import matplotlib.pyplot as plt
        def _save_plt_fig(path):
            plt.savefig(path,bbox_inches='tight', pad_inches =0.0,format='pdf')
            plt.close()

        save_function_dict[FileDatatypes.PDF] = _save_plt_fig
        file_extension_dict[FileDatatypes.PDF] = 'pdf'

    if (FileDatatypes.SHELL.value in supported_datatypes) or can_load_all:
        def _save_shell_script(data,path):
            with open(path, 'w') as f:
                f.write(data)
            os.chmod(path, 0o755)
        
        save_function_dict[FileDatatypes.SHELL] = _save_shell_script
        file_extension_dict[FileDatatypes.SHELL] = 'sh'

    if (FileDatatypes.SLURM_SCRIPT.value in supported_datatypes) or can_load_all:
        def _save_slurm_script(data,path):
            with open(path, 'w') as f:
                f.write(data)
            os.chmod(path, 0o755)
        
        save_function_dict[FileDatatypes.SLURM_SCRIPT] = _save_slurm_script
        file_extension_dict[FileDatatypes.SLURM_SCRIPT] = 'sh'

    if (FileDatatypes.SLURM_OUTPUT_STD.value in supported_datatypes) or can_load_all:
        file_extension_dict[FileDatatypes.SLURM_OUTPUT_STD] = 'out'

    if (FileDatatypes.SLURM_OUTPUT_ERR.value in supported_datatypes) or can_load_all:
        file_extension_dict[FileDatatypes.SLURM_OUTPUT_ERR] = 'err'

    return load_function_dict,save_function_dict,file_extension_dict


from setuptools import setup, find_packages

setup(
    name='file_golem',
    version='0.1.8',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'omegaconf',
        'filelock'
    ],
    # entry_points={
    #     'console_scripts': [
    #         'file_io_main = file_io.main:main',
    #     ],
    # },
    author='Cameron Braunstein',
    author_email='your.email@example.com',
    description='A file i/o utility to handle complex projects with multiple datasets, project locations and calls to 3rd party codebases.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/CameronBraunstein/file_golem',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)